package wwc.messaging;

import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.net.Socket;
import java.util.Hashtable;


import salsa.language.*;
import salsa.language.exceptions.*;
import salsa.naming.*;
import salsa.messaging.*;

/**
 * This class is the reference implementation of the MessagingService. The
 * service is implemented as a RMSP client.
 * 
 * This implementation uses the NamingService interface and the salsa.language.
 * ServiceFactory for UAN resolution.
 * 
 * @author stepha
 */
public class PersistentMessenger implements TransportService {
	/**
	 * This is the maximum number of times a message maybe forwarded before
	 * the name server is queried. Set by system property -DforwardThreshold=<n>
	 * The default value is 4.
	 */
	private int forwarderThreshold;
	private static Hashtable streamRegistry;

	/**
	 * Constructor for Messenger.
	 */
	public PersistentMessenger() {
		super();
		
		String thresh = System.getProperty( "forwardThreshold" );
		if (thresh != null) forwarderThreshold = Integer.parseInt( thresh );
		else forwarderThreshold = 4;

		streamRegistry = new Hashtable();
	}

	private String entry(UAL ual) {
		String entryString = ual.getProtocol() + "://" + ual.getHost() + ":" + ual.getPort();
		return entryString;
	}

	public synchronized ObjectOutputStream getOutputStream(UAL target) {
		ObjectOutputStream outStream = (ObjectOutputStream)streamRegistry.get( entry(target) );

		if (outStream == null) {
			try {
				Socket outgoing = new Socket(target.getHost(), target.getPort());
				outStream = new ObjectOutputStream(outgoing.getOutputStream());
				streamRegistry.put( entry(target), outStream );
			} catch (Exception e) {
				throw new SalsaWrapperException("Error creating outputstream: ", e);
			}	
		}

		return outStream;
	}

	public void migrate(Actor state, UAL target) {
		Socket outgoing = null;
		try {
			outgoing = new Socket(target.getHost(), target.getPort());
			ObjectOutputStream outStream = new ObjectOutputStream(outgoing.getOutputStream());
			outStream.writeObject( state );
			outStream.flush();
		} catch (IOException e) {
			throw new SalsaWrapperException("Error migrating actor: ", e);
		}

		// Wait for the Acknowledgement that the actor is situated on the remote theater.
		try {
			RMSPAcknowledgement ack = (RMSPAcknowledgement) new ObjectInputStream( outgoing.getInputStream() ).readObject();
		} catch (Exception e) {
			throw new SalsaWrapperException( "Did not recieve RmspAcknowledgement. ", e);
		}

		try {
			outgoing.close();
		} catch (IOException e) {
			throw new SalsaWrapperException("Error migrating actor: ", e);
		}
	}

	public void send(Message message, UAL target) {
		// Determine the destination of this message.
		if ((message.getForwarded() >= forwarderThreshold) || (target == null)) {
			UAN uan = message.getTarget().getUAN();

			// Resolve the location.
			target = ServiceFactory.getNaming().resolve( uan );
			message.setForwarded( 0 );

			ServiceFactory.getTheater().setEntry(uan, target);
		}

		// Update the number of times the message has been forwarded.
		message.setForwarded( message.getForwarded() + 1 );

		try {
			ObjectOutputStream outStream = getOutputStream(target);
			outStream.writeObject( message );
			outStream.flush();
		} catch (IOException e) {
			throw new SalsaWrapperException("Error sending message: ", e);
		}
	}
}
